
(require 'cl)

(defcustom tab-bar-tab-width-max 220
  "Maximum number of pixels allowed for the width of a tab name.
When nil, there is no limit on maximum width."
  :type '(choice (const :tag "No limit" nil)
                 (integer :tag "Max width" :value 220))
  :group 'tab-bar
  :version "29.1")

(defun tab-bar-format-list-pixelize (orig-fun &rest args)
  (let ((non-tab 0) ;; all non-tab items including buttons and indicators
        (num-tab 0) ;; number of adjustable tabs
        (tab-width tab-bar-tab-width-max)
        (items (apply orig-fun args)))
    (dolist (item items)
      (when (and (eq (nth 1 item) 'menu-item)
                 (stringp (nth 2 item)))
        (let ((width (string-pixel-width (propertize (nth 2 item) 'face 'tab-bar))))
          (if (or (not (memq (nth 3 item) '(ignore tab-bar-menu-bar)))
                  (eq (nth 0 item) 'current-tab))
              (setq num-tab (+ num-tab 1))
            (if (not (eq (nth 0 item) 'align-right))
                (setq non-tab (+ non-tab width))
              (setq non-tab (+ non-tab 100)))))))
    (when (> num-tab 0)
      (setq tab-width (min (/ (- (frame-pixel-width) non-tab) num-tab) tab-bar-tab-width-max)))
    (dolist (item items)
      (when (and (eq (nth 1 item) 'menu-item)
                 (stringp (nth 2 item))
                 (or (and (not (memq (nth 3 item) '(ignore tab-bar-menu-bar)))
                          (not (memq (nth 0 item) '(add-tab history-back history-forward))))
                     (eq (nth 0 item) 'current-tab)))
        (let* ((width (string-pixel-width (propertize (nth 2 item) 'face 'tab-bar)))
               (name (nth 2 item))
               (len (length name))
               (ins-pos (- len (if (get-text-property (1- len) 'close-tab name) 1 0)))
               del-pos)
          (while (< (string-pixel-width (propertize (nth 2 item) 'face 'tab-bar)) tab-width)
            (setf (substring (nth 2 item) ins-pos ins-pos)
                  (apply 'propertize " " (text-properties-at 0 (nth 2 item)))))
          (while (> (string-pixel-width (propertize (nth 2 item) 'face 'tab-bar)) tab-width)
            (setq name (nth 2 item) len (length name)
                  del-pos (- len (if (get-text-property (1- len) 'close-tab name) 1 0)))
            (setf (substring (nth 2 item) (1- del-pos) del-pos) "")))))
    items))

(advice-add 'tab-bar-format-list :around
            'tab-bar-format-list-pixelize
            '((name . tab-bar-format-list-pixelize)))
